# May-03-2024
# utils_directory.py

from pathlib import Path


def init_directory(dir_name):

    dir_path = Path.cwd() / dir_name

    if dir_path.is_dir():
        for child in dir_path.glob('*'):
            if child.is_file():
                child.unlink()
    else:
        create_directory(dir_name)


def create_directory(dir_name):

    dir_path = Path.cwd() / dir_name

    if not dir_path.is_dir():
        dir_path.mkdir()


def remove_directory(dir_name):

    dir_path = Path.cwd() / dir_name

    if dir_path.is_dir():
        for child in dir_path.glob('*'):
            if child.is_file():
                child.unlink()
        dir_path.rmdir()



# May-03-2024
# shell_sort.py


# vect1: 1, 2, 3, ...
def shell_sort_increase(vect1, vect2, n):

    gap = n // 2

    while gap > 0:

        for i in range(gap, n):

            j = i - gap

            while j >= 0 and vect1[j] > vect1[j+gap]:

                temp = vect1[j]
                vect1[j] = vect1[j+gap]
                vect1[j+gap] = temp

                temp = vect2[j]
                vect2[j] = vect2[j+gap]
                vect2[j+gap] = temp

                j = j - gap

        gap = gap // 2


# vect1: 3, 2, 1, ...
def shell_sort_decrease(vect1, vect2, n):

    gap = n // 2

    while gap > 0:

        for i in range(gap, n):

            j = i - gap

            while j >= 0 and vect1[j] < vect1[j + gap]:
                temp = vect1[j]
                vect1[j] = vect1[j + gap]
                vect1[j + gap] = temp

                temp = vect2[j]
                vect2[j] = vect2[j + gap]
                vect2[j + gap] = temp

                j = j - gap

        gap = gap // 2



# May-03-2024
# get_similarity.py

import sys
import cv2 as cv

from mnist_separator.src.to_canonical import to_canonical
from mnist_separator.src.calc_similarity import calc_similarity


def get_similarity(path_1: str, path_2: str) -> float:

    image1_mnist = cv.imread(path_1, cv.IMREAD_UNCHANGED)
    if image1_mnist is None:
        print(f'\nERROR: Unable to read {path_1}.')
        sys.exit(1)

    image2_mnist = cv.imread(path_2, cv.IMREAD_UNCHANGED)
    if image2_mnist is None:
        print(f'\nERROR: Unable to read {path_2}.')
        sys.exit(1)

    image1_canon = to_canonical(image1_mnist)
    image2_canon = to_canonical(image2_mnist)

    similarity = calc_similarity(image1_canon, image2_canon)

    return similarity



# May-03-2024
# calc_similarity_next.py

import numpy as np

from mnist_separator.src import cfg


def calc_similarity_next(
        magnitude_1: np.float32,
        magnitude_2: np.float32) -> float:

    arr_zero = np.zeros(cfg.dsize_roi, dtype=np.float32)
    arr_1 = np.maximum(magnitude_1, arr_zero)
    arr_2 = np.maximum(magnitude_2, arr_zero)

    arr_min = np.minimum(arr_1, arr_2)
    arr_max = np.maximum(arr_1, arr_2)

    similarity_map = np.subtract(arr_max, arr_min, where=arr_min > 0)
    nonzero_count = np.count_nonzero(similarity_map)

    if nonzero_count == 0:
        similarity = 1.0
    else:
        average = np.sum(similarity_map) / nonzero_count
        similarity = 1.0 / (1.0 + cfg.param_similarity * average)

    return similarity



# May-03-2024
# get_similarity_group.py

import os
import cv2 as cv
import numpy as np

from mnist_separator.src.calc_similarity import calc_similarity


def get_similarity_group(path_image_test: str, path_train: str) -> float:
    """
    Нахождение максимального значения степени сходства между
    заданным каноническим изображением и всеми каноническими
    изображениями группы.
    """

    similarity_group: float = 0

    list_train = list_visible_files(path_train)
    if len(list_train) == 0:
        return similarity_group

    image_canon = cv.imread(path_image_test, cv.IMREAD_UNCHANGED)

    for fname in list_train:

        path_image_train = path_train + '/' + fname
        image_train_canon = cv.imread(path_image_train, cv.IMREAD_UNCHANGED)

        similarity = calc_similarity(np.uint8(image_canon), np.uint8(image_train_canon))

        similarity_group = max(similarity, similarity_group)

    return similarity_group


def list_visible_files(dir_name):
    visible_files = []
    for file in os.listdir(dir_name):
        if not file.startswith('.'):
            visible_files.append(file)
    return visible_files



# May-03-2024
# calc_similarity.py

import numpy as np

from mnist_separator.src.magnitude import get_magn_roi, get_magn_half
from mnist_separator.src.get_peaks import get_peaks
from mnist_separator.src.sort_peaks import sort_peaks
from mnist_separator.src.utils_list_of_peaks import cut_list_of_peaks
from mnist_separator.src.calc_similarity_2pnts import calc_similarity_2pnts
from mnist_separator.src.calc_similarity_3pnts import calc_similarity_3pnts


def calc_similarity(image: np.uint8, templ: np.uint8) -> float:

    # Image
    # ---------------------------------------------------------
    image_magn_roi = get_magn_roi(image)
    image_magn_half = get_magn_half(image_magn_roi)

    image_list_of_peaks = get_peaks(image_magn_half)
    image_list_of_peaks_sort = sort_peaks(image_magn_half, image_list_of_peaks)
    image_list_of_peaks_calc = cut_list_of_peaks(image_list_of_peaks_sort)
    # ---------------------------------------------------------

    # Templ
    # ---------------------------------------------------------
    templ_magn_roi = get_magn_roi(templ)
    templ_magn_half = get_magn_half(templ_magn_roi)

    templ_list_of_peaks = get_peaks(templ_magn_half)
    templ_list_of_peaks_sort = sort_peaks(templ_magn_half, templ_list_of_peaks)
    templ_list_of_peaks_calc = cut_list_of_peaks(templ_list_of_peaks_sort)
    # ---------------------------------------------------------

    # ---------------------------------------------------------
    similarity_result_2pnts = calc_similarity_2pnts(
                                        image_magn_roi, templ_magn_roi,
                                        image_list_of_peaks_calc,
                                        templ_list_of_peaks_calc)

    similarity_result_3pnts = calc_similarity_3pnts(
                                        image_magn_roi, templ_magn_roi,
                                        image_list_of_peaks_calc,
                                        templ_list_of_peaks_calc)
    # ---------------------------------------------------------

    return max(similarity_result_2pnts, similarity_result_3pnts)



# May-03-2024
# utils_list_of_peaks.py

from mnist_separator.src import cfg


def cut_list_of_peaks(list_of_peaks) -> list:

    # remove first peak
    list_of_peaks_1 = list_of_peaks[1:]

    length = len(list_of_peaks_1)

    if length > cfg.n_peaks:
        # remove extra peaks from the end
        list_of_peaks_1 = list_of_peaks_1[:cfg.n_peaks]

    return list_of_peaks_1


def save_list_of_peaks(tag, directory, magn, list_of_peaks):

    data_path = directory + '/' + tag + '.txt'

    with open(data_path, 'w') as fp:
        n: int = 0
        for item in list_of_peaks:
            y = item[0]
            x = item[1]
            fp.write(f'{n}:\t {item}\t magn = {magn[y, x]}\n')
            n += 1


def save_list_of_peaks_1(tag, directory, list_of_peaks):

    data_path = directory + '/' + tag + '.txt'

    with open(data_path, 'w') as fp:
        n: int = 0
        for item in list_of_peaks:
            fp.write(f'{n}:\t {item}\n')
            n += 1


def read_list_of_peaks_1(tag, directory):

    list_of_peaks = []

    data_path = directory + '/' + tag + '.txt'

    with open(data_path, 'r') as fp:
        for line in fp:

            # remove linebreak from a current item
            item = line[:-1]

            list_of_peaks.append(item)

    return list_of_peaks



# May-03-2024
# timer.py

import time
from pathlib import Path
from datetime import date, datetime

global timer_start


def init_timer():
    global timer_start
    timer_start = time.time()


def save_elapsed_time_hour_min_sec(dir_name: str):
    global timer_start

    today = date.today()
    current_date = today.strftime("%B %d, %Y")
    c = datetime.now()
    current_time = c.strftime('%H:%M:%S')
    when = current_date + '\t' + current_time

    timer_end = time.time()
    seconds = int(timer_end - timer_start)
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)

    fname = Path.cwd() / dir_name / "date_time.txt"
    with open(fname, 'w') as fname:
        str_time = f'{h:02d} hour(s)  {m:02d} min  {s:02d} sec'
        fname.write(when + '\n\n')
        fname.write('Elapsed time = ' + str_time + '\n')



# May-03-2024
# save_result.py

from mnist_separator.src import glbl
from mnist_separator.src.utils_list import save_list


def save_result(list_result_0, list_result_1,
                list_result_2, list_result_3,
                list_result_4, list_result_5,
                list_result_6, list_result_7,
                list_result_8, list_result_9):

    if len(list_result_0) > 0:
        save_list('list_result_0.txt', glbl.result_name, list_result_0)
    if len(list_result_1) > 0:
        save_list('list_result_1.txt', glbl.result_name, list_result_1)
    if len(list_result_2) > 0:
        save_list('list_result_2.txt', glbl.result_name, list_result_2)
    if len(list_result_3) > 0:
        save_list('list_result_3.txt', glbl.result_name, list_result_3)
    if len(list_result_4) > 0:
        save_list('list_result_4.txt', glbl.result_name, list_result_4)
    if len(list_result_5) > 0:
        save_list('list_result_5.txt', glbl.result_name, list_result_5)
    if len(list_result_6) > 0:
        save_list('list_result_6.txt', glbl.result_name, list_result_6)
    if len(list_result_7) > 0:
        save_list('list_result_7.txt', glbl.result_name, list_result_7)
    if len(list_result_8) > 0:
        save_list('list_result_8.txt', glbl.result_name, list_result_8)
    if len(list_result_9) > 0:
        save_list('list_result_9.txt', glbl.result_name, list_result_9)



# May-03-2024
# calc_similarity_3pnts.py

import cv2 as cv
import numpy as np

from mnist_separator.src import cfg
from mnist_separator.src.calc_similarity_next import calc_similarity_next
from mnist_separator.src.get_angle import get_angle_axe_x


def calc_similarity_3pnts(
                image_magn_roi, templ_magn_roi,
                image_list_of_peaks_calc,
                templ_list_of_peaks_calc) -> float:

    image_table, n_image_table_rows = create_table_3pnts(image_list_of_peaks_calc)
    templ_table, n_templ_table_rows = create_table_3pnts(templ_list_of_peaks_calc)

    similarity_result: float = 0
    for j in range(n_image_table_rows):

        angle_image: int = image_table[j, 4]

        if angle_image > 0:
            y1_image = image_table[j, 0]
            x1_image = image_table[j, 1]
            y2_image = image_table[j, 2]
            x2_image = image_table[j, 3]
        else:
            y1_image = image_table[j, 2]
            x1_image = image_table[j, 3]
            y2_image = image_table[j, 0]
            x2_image = image_table[j, 1]

        for i in range(n_templ_table_rows):

            angle_templ: int = templ_table[i, 4]

            if angle_templ > 0:
                y1_templ = templ_table[i, 0]
                x1_templ = templ_table[i, 1]
                y2_templ = templ_table[i, 2]
                x2_templ = templ_table[i, 3]
            else:
                y1_templ = templ_table[i, 2]
                x1_templ = templ_table[i, 3]
                y2_templ = templ_table[i, 0]
                x2_templ = templ_table[i, 1]

            # angle filter
            angle_image = abs(angle_image)
            angle_templ = abs(angle_templ)
            if angle_image > angle_templ:
                angle_ratio: float = float(angle_image) / float(angle_templ)
            else:
                angle_ratio: float = float(angle_templ) / float(angle_image)

            if angle_ratio > cfg.angle_ratio_threshold:
                continue

            pts_image = np.float32([[x1_image, y1_image],
                                    [x2_image, y2_image],
                                    [cfg.X0, cfg.Y0]])

            pts_templ = np.float32([[x1_templ, y1_templ],
                                    [x2_templ, y2_templ],
                                    [cfg.X0, cfg.Y0]])

            mat_affine = cv.getAffineTransform(pts_image, pts_templ)

            image_magn_roi_warp \
                = cv.warpAffine(image_magn_roi, mat_affine, cfg.dsize_roi)

            similarity_current \
                = calc_similarity_next(np.float32(image_magn_roi_warp),
                                       np.float32(templ_magn_roi))

            if similarity_current > similarity_result:
                similarity_result = similarity_current

    return similarity_result


def create_table_3pnts(list_of_peaks) -> (np.ndarray, int):

    n_peaks: int = len(list_of_peaks)

    first_table = np.zeros((n_peaks, 2), dtype=np.int32)

    for n in range(n_peaks):
        peak = list_of_peaks[n]
        first_table[n, 0] = peak[0]  # y пикa
        first_table[n, 1] = peak[1]  # x пикa

    n_second_table: int = (n_peaks * n_peaks - n_peaks) // 2

    second_table = np.zeros((n_second_table, 5), dtype=np.int32)

    n_rows: int = 0
    for j in range(n_peaks - 1):

        y1 = int(first_table[j, 0])
        x1 = int(first_table[j, 1])
        angle_1: float = get_angle_axe_x(x1, y1)

        for i in range(j, n_peaks):

            y2 = int(first_table[i, 0])
            x2 = int(first_table[i, 1])
            angle_2: float = get_angle_axe_x(x2, y2)

            peaks_angle: int = int(angle_2 - angle_1)

            peaks_angle_abs: int = abs(peaks_angle)
            if ((peaks_angle_abs < cfg.angle_min) or
                    (peaks_angle_abs > cfg.angle_max)):
                continue

            second_table[n_rows, 0] = y1
            second_table[n_rows, 1] = x1
            second_table[n_rows, 2] = y2
            second_table[n_rows, 3] = x2
            second_table[n_rows, 4] = peaks_angle

            n_rows += 1

    return second_table, n_rows



# May-03-2024
# utils_table.py

from numpy import ndarray


def save_table(tag, directory, table: ndarray, n_rows: int, n_cols: int):

    data_path = directory + '/' + tag + '.txt'

    with open(data_path, 'w') as fp:

        for j in range(n_rows):
            # fp.write(f'{j}:\t\t')
            for i in range(n_cols):
                fp.write(f'{table[j, i]}\t\t')
            fp.write(f'\n')



# May-03-2024
# load_data.py

import sys
import os
import shutil
from pathlib import Path


def load_data(dir_mnist, digit, num_images):

    if dir_mnist.exists() and dir_mnist.is_dir():

        dir_src = dir_mnist / 'data_test' / str(digit)

        dir_dst = Path.cwd() / '$mnist' / 'test'

        # Get a list of files in the source directory
        files = os.listdir(dir_src)

        # Sort the files based on their creation time or modification time
        files.sort(key=lambda x: os.path.getmtime(os.path.join(dir_src, x)))

        # Copy each file from the source directory to the destination directory
        n = 0
        for file in files:
            src_file_path = os.path.join(dir_src, file)
            dst_file_path = os.path.join(dir_dst, file)
            shutil.copy(src_file_path, dst_file_path)
            n += 1
            if n == num_images:
                break
    else:
        sys.exit(f"\nDirectory {dir_mnist} does not exist.")



# May-03-2024
# to_canonical.py

import cv2 as cv
import numpy as np


def to_canonical(image_in):

    width = 128
    height = 128

    # dim = (width, height)
    image_resize = cv.resize(image_in, (width, height), cv.INTER_LANCZOS4)

    threshold, image_thr \
        = cv.threshold(image_resize, 0, 255,
                       cv.THRESH_BINARY + cv.THRESH_OTSU)

    image_correlation = get_correlation(image_thr)

    threshold, image_out \
        = cv.threshold(image_correlation, 0, 255,
                       cv.THRESH_BINARY + cv.THRESH_OTSU)

    return image_out


def get_correlation(scene):

    # Calc DFT dimension.
    # ---------------------------------------------------------
    scene_rows = scene.shape[0]
    scene_cols = scene.shape[1]

    height_dft = 256
    width_dft = 256
    dft_shape = (height_dft, width_dft)
    # ---------------------------------------------------------

    # Scene
    # ---------------------------------------------------------
    scene_array_dft = np.zeros(dft_shape, dtype=np.float32)

    scene_array_dft[0:scene_rows, 0:scene_cols] = scene[0::, 0::]

    scene_dft = cv.dft(scene_array_dft, flags=cv.DFT_COMPLEX_OUTPUT)
    # ---------------------------------------------------------

    # Pattern
    # ---------------------------------------------------------
    pattern_size: int = 1
    pattern_rows = pattern_size
    pattern_cols = pattern_size
    pattern_shape = (pattern_rows, pattern_cols)
    pattern = np.empty(pattern_shape, dtype=np.uint8)
    pattern.fill(255)
    # ---------------------------------------------------------
    pattern_array_dft = np.zeros(dft_shape, dtype=np.float32)

    pattern_array_dft[
                height_dft - pattern_rows:height_dft:,
                width_dft - pattern_cols:width_dft:] = pattern[0::, 0::]

    pattern_dft = cv.dft(pattern_array_dft, flags=cv.DFT_COMPLEX_OUTPUT)
    # ---------------------------------------------------------

    # Product 1
    # ---------------------------------------------------------
    product1_dft = np.zeros((height_dft, width_dft, 2), dtype=np.float32)

    dft_mul(scene_dft, pattern_dft, product1_dft)
    # ---------------------------------------------------------

    #  Laplace filtering
    # ---------------------------------------------------------
    array_dft = laplace(height_dft, width_dft)

    filter_dft = cv.dft(array_dft, flags=cv.DFT_COMPLEX_OUTPUT)
    # ---------------------------------------------------------

    # Product 2
    # ---------------------------------------------------------
    product2_dft = np.zeros((height_dft, width_dft, 2), dtype=np.float32)

    dft_mul(product1_dft, filter_dft, product2_dft)
    # ---------------------------------------------------------

    # Inverse product2_dft.
    # ---------------------------------------------------------
    inverse_product2_dft = cv.idft(product2_dft)

    re = inverse_product2_dft[:, :, 0]

    re[re < 0.0] = 0.0
    # ---------------------------------------------------------

    # Get Correlation.
    # ---------------------------------------------------------
    correlation = np.zeros((scene_rows, scene_cols), dtype='float32')
    correlation_norm = np.zeros((scene_rows, scene_cols), dtype=np.float32)

    correlation[0::, 0::] = re[0:scene_rows:, 0:scene_cols:]

    cv.normalize(correlation, correlation_norm, 0, 255, cv.NORM_MINMAX)
    # ---------------------------------------------------------

    return np.uint8(correlation_norm)


def dft_mul(src1, src2, dst):

    x1 = src1[:, :, 0]
    y1 = src1[:, :, 1]

    x2 = src2[:, :, 0]
    y2 = src2[:, :, 1]

    dst[:, :, 0] = (x1 * x2) - (y1 * y2)
    dst[:, :, 1] = (x1 * y2) + (x2 * y1)


def laplace(height_dft, width_dft):

    array_filter = np.zeros((height_dft, width_dft), dtype='float32')

    array_filter[0, 0] = -1
    array_filter[0, 1] = -1
    array_filter[0, 2] = -1

    array_filter[1, 0] = -1
    array_filter[1, 1] = 8
    array_filter[1, 2] = -1

    array_filter[2, 0] = -1
    array_filter[2, 1] = -1
    array_filter[2, 2] = -1

    return array_filter



# May-03-2024
# glbl.py


result_name: str = '_RESULT_'

path_test_mnist: str = ''
path_train_0_mnist: str = ''
path_train_1_mnist: str = ''
path_train_2_mnist: str = ''
path_train_3_mnist: str = ''
path_train_4_mnist: str = ''
path_train_5_mnist: str = ''
path_train_6_mnist: str = ''
path_train_7_mnist: str = ''
path_train_8_mnist: str = ''
path_train_9_mnist: str = ''

path_test_canon: str = ''
path_train_0_canon: str = ''
path_train_1_canon: str = ''
path_train_2_canon: str = ''
path_train_3_canon: str = ''
path_train_4_canon: str = ''
path_train_5_canon: str = ''
path_train_6_canon: str = ''
path_train_7_canon: str = ''
path_train_8_canon: str = ''
path_train_9_canon: str = ''



# May-03-2024
# cfg.py

path_image: str = ''
path_templ: str = ''

# Число локальных максимумов в
# верхней полуплоскости.
n_peaks: int = 9

size_dft: int = 2048
lp_param: int = 4
size_roi: int = size_dft // lp_param
dsize_roi = (size_roi, size_roi)
size_roi_half: int = size_roi // 2
X0: int = size_roi_half
Y0: int = size_roi_half
center = (X0, Y0)

angle_min: int = 20
angle_max: int = 180 - angle_min
angle_ratio_threshold: float = 3.0

param_similarity: float = 0.07

y1_image: int = -1
x1_image: int = -1
y2_image: int = -1
x2_image: int = -1

y1_templ: int = -1
x1_templ: int = -1
y2_templ: int = -1
x2_templ: int = -1



# May-03-2024
# sort_peaks.py

import numpy as np

from mnist_separator.src.shell_sort import shell_sort_decrease


def sort_peaks(magn, list_in):

    len_list = len(list_in)
    arr_magn = np.zeros(len_list, dtype=np.float32)

    n = 0
    for item in list_in:
        y = int(item[0])
        x = int(item[1])

        arr_magn[n] = magn[y, x]
        n += 1

    # Сортировка arr_magn.
    order = np.arange(len_list)
    shell_sort_decrease(arr_magn, order, len_list)

    # Пики в списке list_out отсортированы по значению магнитуды.
    list_out = []
    for i in range(len_list):
        j = order[i]
        item = list_in[j]
        list_out.append(item)

    return list_out



# May-03-2024
# go.py

import sys
import os
import numpy as np
import random
from tqdm import tqdm

from mnist_separator.src import glbl
from mnist_separator.src.timer import init_timer, save_elapsed_time_hour_min_sec
from mnist_separator.src.convert_data import convert_data
from mnist_separator.src.get_similarity_group import get_similarity_group
from mnist_separator.src.utils_list import list_visible_files
from mnist_separator.src.save_result import save_result


def go():

    init_timer()

    convert_data()

    list_result_0 = []
    list_result_1 = []
    list_result_2 = []
    list_result_3 = []
    list_result_4 = []
    list_result_5 = []
    list_result_6 = []
    list_result_7 = []
    list_result_8 = []
    list_result_9 = []

    list_test_files = []
    list_test_files_conon = list_visible_files(glbl.path_test_canon)
    for fname in list_test_files_conon:
        path_test_canon = glbl.path_test_canon + '/' + fname
        list_test_files.append(path_test_canon)

    random.shuffle(list_test_files)
    n_test_files = len(list_test_files)
    if n_test_files == 0:
        sys.exit(f'\nNo data')

    print()
    for n in tqdm(range(n_test_files), desc='separation'):

        path_test_canon = list_test_files[n]

        list_similarities = [get_similarity_group(path_test_canon, glbl.path_train_0_canon),
                             get_similarity_group(path_test_canon, glbl.path_train_1_canon),
                             get_similarity_group(path_test_canon, glbl.path_train_2_canon),
                             get_similarity_group(path_test_canon, glbl.path_train_3_canon),
                             get_similarity_group(path_test_canon, glbl.path_train_4_canon),
                             get_similarity_group(path_test_canon, glbl.path_train_5_canon),
                             get_similarity_group(path_test_canon, glbl.path_train_6_canon),
                             get_similarity_group(path_test_canon, glbl.path_train_7_canon),
                             get_similarity_group(path_test_canon, glbl.path_train_8_canon),
                             get_similarity_group(path_test_canon, glbl.path_train_9_canon)]

        name_test = os.path.basename(path_test_canon)
        index = np.argmax(list_similarities)

        match index:
            case 0: list_result_0.append(name_test)
            case 1: list_result_1.append(name_test)
            case 2: list_result_2.append(name_test)
            case 3: list_result_3.append(name_test)
            case 4: list_result_4.append(name_test)
            case 5: list_result_5.append(name_test)
            case 6: list_result_6.append(name_test)
            case 7: list_result_7.append(name_test)
            case 8: list_result_8.append(name_test)
            case 9: list_result_9.append(name_test)

    save_result(list_result_0, list_result_1,
                list_result_2, list_result_3,
                list_result_4, list_result_5,
                list_result_6, list_result_7,
                list_result_8, list_result_9)

    save_elapsed_time_hour_min_sec(glbl.result_name)



# May-03-2024
# utils_list.py

import os
from pathlib import Path


def save_list(list_name, dir_name, any_list):

    fname = Path.cwd() / dir_name / list_name

    with open(fname, 'w') as fname:
        for item in any_list:
            fname.write(f'{item}\n')


def print_list(tag, any_list):
    print(f'{tag}:')
    n = 0
    for item in any_list:
        print(f'{n}:\t {item}')
        n += 1


def list_visible_files(dir_name):
    visible_files = []
    for file in os.listdir(dir_name):
        if not file.startswith('.'):
            visible_files.append(file)
    return visible_files



# May-03-2024
# get_peaks.py

import numpy as np
from numpy import ndarray

from mnist_separator.src import cfg


color_black: np.uint8 = np.uint8(0)
color_dark_gray: np.uint8 = np.uint8(80)
color_light_gray: np.uint8 = np.uint8(160)

coeff_1: float = 0.03
coeff_2: float = 0.02
coeff_3: float = 0.01

seed_y: int = 0
seed_x: int = 0

peak_y: int = 0
peak_x: int = 0


def get_peaks(magn_half: ndarray):

    global seed_y, seed_x, peak_y, peak_x
    global coeff_1, coeff_2, coeff_3
    global color_black, color_dark_gray, color_light_gray

    z_max = np.amax(magn_half)
    z_step = coeff_1 * z_max

    n_detected_peaks = 0

    list_of_peaks = []

    z_cut = z_max
    n_cut = 1

    while True:

        z_cut -= z_step

        if z_cut < 0.0:
            break

        canvas = np.where(magn_half > z_cut,
                          color_light_gray, color_black)

        while True:

            result = np.where(canvas == color_light_gray)

            number = result[0].size
            if number == 0:
                break

            list_coord = list(zip(result[0], result[1]))

            first = list_coord[0]
            seed_y = int(first[0])
            seed_x = int(first[1])

            fill_canvas(canvas,
                        magn_half,
                        color_light_gray,
                        color_dark_gray)

            if list_of_peaks.count((peak_y, peak_x)) == 0:
                list_of_peaks.append((peak_y, peak_x))

            n_detected_peaks = len(list_of_peaks)

        if n_detected_peaks >= (cfg.n_peaks + 1):
            # print(f'n_detected_peaks: {n_detected_peaks}')
            break

        if n_detected_peaks > 3:
            z_step = coeff_2 * z_max
        if n_detected_peaks > 5:
            z_step = coeff_3 * z_max

        n_cut += 1

    return list_of_peaks


def fill_canvas(
        canvas,
        magnitude: ndarray,
        old_value: np.uint8,
        new_value: np.uint8):

    """
        https://gist.github.com/JDWarner/1158a9515c7f1b1c21f1
        Thanks, JD!
    """

    global seed_y, seed_x, peak_y, peak_x

    ysize_1 = canvas.shape[0] - 1
    xsize_1 = canvas.shape[1] - 1

    stack = {(seed_y, seed_x)}

    magnitude_max = np.float32(0)

    while stack:

        y, x = stack.pop()

        temp = magnitude[y, x]

        if temp > magnitude_max:

            magnitude_max = temp

            if y > 0 and (0 < x < xsize_1):
                peak_y = y
                peak_x = x

        if canvas[y, x] == old_value:

            canvas[y, x] = new_value

            if y > 0:
                stack.add((y - 1, x))
            if y < ysize_1:
                stack.add((y + 1, x))
            if x > 0:
                stack.add((y, x - 1))
            if x < xsize_1:
                stack.add((y, x + 1))



# May-03-2024
# convert_data.py

import cv2 as cv
import os
from pathlib import Path

from mnist_separator.src import glbl
from mnist_separator.src.to_canonical import to_canonical


glbl.path_test_mnist = str(Path.cwd() / '$mnist' / 'test')
glbl.path_train_0_mnist = str(Path.cwd() / '$mnist' / 'train_0')
glbl.path_train_1_mnist = str(Path.cwd() / '$mnist' / 'train_1')
glbl.path_train_2_mnist = str(Path.cwd() / '$mnist' / 'train_2')
glbl.path_train_3_mnist = str(Path.cwd() / '$mnist' / 'train_3')
glbl.path_train_4_mnist = str(Path.cwd() / '$mnist' / 'train_4')
glbl.path_train_5_mnist = str(Path.cwd() / '$mnist' / 'train_5')
glbl.path_train_6_mnist = str(Path.cwd() / '$mnist' / 'train_6')
glbl.path_train_7_mnist = str(Path.cwd() / '$mnist' / 'train_7')
glbl.path_train_8_mnist = str(Path.cwd() / '$mnist' / 'train_8')
glbl.path_train_9_mnist = str(Path.cwd() / '$mnist' / 'train_9')

glbl.path_test_canon = str(Path.cwd() / '$canon' / 'test')
glbl.path_train_0_canon = str(Path.cwd() / '$canon' / 'train_0')
glbl.path_train_1_canon = str(Path.cwd() / '$canon' / 'train_1')
glbl.path_train_2_canon = str(Path.cwd() / '$canon' / 'train_2')
glbl.path_train_3_canon = str(Path.cwd() / '$canon' / 'train_3')
glbl.path_train_4_canon = str(Path.cwd() / '$canon' / 'train_4')
glbl.path_train_5_canon = str(Path.cwd() / '$canon' / 'train_5')
glbl.path_train_6_canon = str(Path.cwd() / '$canon' / 'train_6')
glbl.path_train_7_canon = str(Path.cwd() / '$canon' / 'train_7')
glbl.path_train_8_canon = str(Path.cwd() / '$canon' / 'train_8')
glbl.path_train_9_canon = str(Path.cwd() / '$canon' / 'train_9')


def convert_data():
    mnist_to_canonical(glbl.path_test_mnist, glbl.path_test_canon)

    mnist_to_canonical(glbl.path_train_0_mnist, glbl.path_train_0_canon)
    mnist_to_canonical(glbl.path_train_1_mnist, glbl.path_train_1_canon)
    mnist_to_canonical(glbl.path_train_2_mnist, glbl.path_train_2_canon)
    mnist_to_canonical(glbl.path_train_3_mnist, glbl.path_train_3_canon)
    mnist_to_canonical(glbl.path_train_4_mnist, glbl.path_train_4_canon)
    mnist_to_canonical(glbl.path_train_5_mnist, glbl.path_train_5_canon)
    mnist_to_canonical(glbl.path_train_6_mnist, glbl.path_train_6_canon)
    mnist_to_canonical(glbl.path_train_7_mnist, glbl.path_train_7_canon)
    mnist_to_canonical(glbl.path_train_8_mnist, glbl.path_train_8_canon)
    mnist_to_canonical(glbl.path_train_9_mnist, glbl.path_train_9_canon)


def mnist_to_canonical(dir_name_in, dir_name_out):

    for fname in os.listdir(dir_name_in):

        if fname.startswith('.'):
            continue

        path_in = dir_name_in + '/' + fname
        path_out = dir_name_out + '/' + fname

        image_mnist = cv.imread(path_in, cv.IMREAD_GRAYSCALE)

        image_canon = to_canonical(image_mnist)

        cv.imwrite(path_out, image_canon)



# May-03-2024
# clean_all.py

from pathlib import Path

from mnist_separator.src import glbl
from mnist_separator.src.utils_directory import init_directory


def clean_all():

    init_directory(glbl.result_name)

    path_mnist = Path.cwd() / '$mnist'
    path_conon = Path.cwd() / '$canon'

    if path_mnist.is_dir():
        rmtree(path_mnist)
    if path_conon.is_dir():
        rmtree(path_conon)


def rmtree(root):

    for p in root.iterdir():
        if p.is_dir():
            rmtree(p)
        else:
            p.unlink()

    root.rmdir()



# May-03-2024
# create_dirs.py

from pathlib import Path
import os


def create_dirs():
    create_mnist_dirs()
    create_canon_dirs()


def create_mnist_dirs():

    dir_name = '$mnist'
    dir_path = Path.cwd() / dir_name

    result = os.path.isdir(dir_path)
    if result:
        return
    else:
        dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'test'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_0'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_1'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_2'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_3'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_4'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_5'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_6'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_7'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_8'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_9'
    dir_path.mkdir()


def create_canon_dirs():

    dir_name = '$canon'
    dir_path = Path.cwd() / dir_name

    result = os.path.isdir(dir_path)
    if result:
        return
    else:
        dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'test'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_0'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_1'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_2'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_3'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_4'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_5'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_6'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_7'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_8'
    dir_path.mkdir()

    dir_path = Path.cwd() / dir_name / 'train_9'
    dir_path.mkdir()



# May-03-2024
# calc_similarity_2pnts.py

import cv2 as cv
import numpy as np
import math

from mnist_separator.src import cfg
from mnist_separator.src.calc_similarity_next import calc_similarity_next
from mnist_separator.src.get_angle import get_angle_axe_x


def calc_similarity_2pnts(
                image_magn_roi, templ_magn_roi,
                image_list_of_peaks_calc,
                templ_list_of_peaks_calc) -> float:

    image_table_2pnts = create_table_2pnts(image_list_of_peaks_calc)
    n_image_peaks = len(image_list_of_peaks_calc)

    templ_table_2pnts = create_table_2pnts(templ_list_of_peaks_calc)
    n_templ_peaks = len(templ_list_of_peaks_calc)

    similarity_result: float = 0
    for j in range(n_image_peaks):

        length_src = image_table_2pnts[j, 2]
        angle_src = image_table_2pnts[j, 3]

        for i in range(n_templ_peaks):

            length_dst = templ_table_2pnts[i, 2]
            angle_dst = templ_table_2pnts[i, 3]

            angle: float = float(angle_dst - angle_src)
            scale: float = float(length_dst / length_src)
            mat_rotate = cv.getRotationMatrix2D(cfg.center, angle, scale)

            image_magn_roi_warp \
                = cv.warpAffine(image_magn_roi, mat_rotate, cfg.dsize_roi)

            similarity_current \
                = calc_similarity_next(np.float32(image_magn_roi_warp),
                                       np.float32(templ_magn_roi))

            if similarity_current > similarity_result:
                similarity_result = similarity_current

    return similarity_result


def create_table_2pnts(list_of_peaks) -> np.ndarray:

    table_rows = len(list_of_peaks)
    table_cols = 4

    table = np.zeros((table_rows, table_cols), dtype=np.float32)

    for n in range(table_rows):

        peak = list_of_peaks[n]
        y_peak = peak[0]
        x_peak = peak[1]

        # coordinates
        table[n, 0] = x_peak
        table[n, 1] = y_peak

        # length
        dx: float = x_peak - cfg.X0
        dy: float = y_peak - cfg.Y0
        table[n, 2] = math.sqrt((dx * dx) + (dy * dy))

        # angle
        table[n, 3] = get_angle_axe_x(x_peak, y_peak)

    return table
